﻿using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.IO;
using System.Linq;
using System.ServiceProcess;
using System.Text;
using System.Threading;

namespace Inedo.Agents.Service
{
    internal sealed class ServerUpdater : IDisposable
    {
        private Dictionary<string, string> stagedDeletes = new Dictionary<string, string>();
        private List<string> installedFiles = new List<string>();
        private StreamWriter logWriter;

        public ServerUpdater(string sourcePath, string targetPath, string serviceName, int serviceId)
        {
            this.SourcePath = sourcePath;
            this.TargetPath = targetPath;
            this.ServiceName = serviceName;
            this.ServiceId = serviceId;

            var logFilePath = Path.Combine(targetPath, "AgentUpdate.log");
            this.logWriter = new StreamWriter(new FileStream(logFilePath, FileMode.Create, FileAccess.Write, FileShare.Read), Encoding.UTF8);
        }

        public string SourcePath { get; }
        public string TargetPath { get; }
        public string ServiceName { get; }
        public int ServiceId { get; }

        public bool Update()
        {
            try
            {
                this.Log($"Attempting to open service process (PID {this.ServiceId})...");
                using (var process = Process.GetProcessById(this.ServiceId))
                {
                    this.Log($"Service process opened; attempting to open service {this.ServiceName}...");

                    using (var service = new ServiceController(this.ServiceName))
                    {
                        this.Log("Service opened.");

                        if (!service.CanStop)
                        {
                            this.Log("Stopping service...");
                            service.Stop();
                            try
                            {
                                service.WaitForStatus(ServiceControllerStatus.Stopped, new TimeSpan(0, 0, 30));
                                this.Log("Service is stopped.");
                            }
                            catch (System.ServiceProcess.TimeoutException)
                            {
                                this.Log("Service has not stopped after 30 seconds.");
                            }
                        }
                        else
                        {
                            this.Log($"Service cannot be stopped in its current state ({service.Status}).");
                        }
                    }

                    if (!process.HasExited)
                    {
                        this.Log("Service process is still running; waiting for it to terminate...");

                        var startTime = DateTime.UtcNow;
                        while (!process.HasExited)
                        {
                            if (DateTime.UtcNow - startTime > new TimeSpan(0, 0, 15))
                            {
                                this.Log("Service process has not terminated after 15 seconds; terminating it forcibly...");
                                process.Kill();
                                this.Log("Process terminated.");
                                break;
                            }

                            Thread.Sleep(500);
                        }
                    }

                    this.Log("Service process exited.");
                }

                this.Log("Getting list of files to update...");

                var currentFiles = from f in Directory.EnumerateFiles(this.TargetPath)
                                   where f.EndsWith(".dll", StringComparison.OrdinalIgnoreCase) || f.EndsWith(".exe", StringComparison.OrdinalIgnoreCase)
                                   select f;

                foreach (var fileName in currentFiles)
                    this.DeleteFile(fileName);

                var updatedFiles = Directory.EnumerateFiles(this.SourcePath);
                foreach (var fileName in updatedFiles)
                    this.InstallFile(fileName);

                this.Log($"Attempting to open service {this.ServiceName}...");

                using (var service = new ServiceController(this.ServiceName))
                {
                    this.Log("Service opened. Starting service...");
                    service.Start();

                    try
                    {
                        service.WaitForStatus(ServiceControllerStatus.Running, new TimeSpan(0, 1, 0));
                    }
                    catch (System.ServiceProcess.TimeoutException)
                    {
                        throw new Exception("Service has not started after 1 minute.");
                    }
                }

                return true;
            }
            catch (Exception ex)
            {
                this.LogError(ex.Message);
                return false;
            }
        }
        public bool Rollback()
        {
            try
            {
                this.Log("Starting rollback.");

                foreach (var fileName in this.installedFiles)
                {
                    int retryCount = 0;
                    this.Log($"Deleting {fileName}...");
                    Retry1:
                    try
                    {
                        File.Delete(fileName);
                    }
                    catch (Exception ex)
                    {
                        if (retryCount < 5)
                        {
                            this.Log($"Could not delete {fileName}; will try again in 5 seconds. Error message: {ex.Message}");
                            Thread.Sleep(5000);
                            retryCount++;
                            goto Retry1;
                        }
                    }
                }

                foreach (var staged in this.stagedDeletes)
                {
                    int retryCount = 0;
                    Retry2:
                    try
                    {
                        File.SetAttributes(staged.Value, File.GetAttributes(staged.Value) & ~FileAttributes.Hidden);
                    }
                    catch
                    {
                    }

                    try
                    {
                        File.Move(staged.Value, staged.Key);
                    }
                    catch (Exception ex)
                    {
                        if (retryCount < 5)
                        {
                            this.Log($"Could not rename {staged.Value} to {staged.Key}; will try again in 5 seconds. Error message: {ex.Message}");
                            Thread.Sleep(5000);
                            retryCount++;
                            goto Retry2;
                        }
                    }
                }

                this.Log($"Opening service {this.ServiceName}...");
                using (var service = new ServiceController(this.ServiceName))
                {
                    this.Log("Service opened. Starting...");
                    service.Start();
                    try
                    {
                        service.WaitForStatus(ServiceControllerStatus.Running, new TimeSpan(0, 1, 0));
                    }
                    catch (System.ServiceProcess.TimeoutException)
                    {
                        throw new Exception("Service has not started after 1 minute.");
                    }
                }

                this.Log("Rollback complete.");
                return true;
            }
            catch (Exception ex)
            {
                this.LogError(ex.Message);
                return false;
            }
        }
        public void Dispose() => this.logWriter?.Dispose();

        private void DeleteFile(string fileName)
        {
            if (!File.Exists(fileName))
                return;

            var tempName = Path.Combine(Path.GetDirectoryName(fileName), Guid.NewGuid().ToString("N"));
            this.Log($"Staging {fileName} for delete as {tempName}...");
            int retryCount = 0;

            Retry:
            try
            {
                File.Move(fileName, tempName);
            }
            catch (Exception ex)
            {
                if (retryCount < 5)
                {
                    this.Log($"Could not rename {fileName}; will try again in 5 seconds. Error message: {ex.Message}");
                    Thread.Sleep(5000);
                    retryCount++;
                    goto Retry;
                }

                throw new Exception($"Could not stage {fileName} for delete. Error message: {ex.Message}");
            }

            this.stagedDeletes.Add(fileName, tempName);

            try
            {
                File.SetAttributes(tempName, File.GetAttributes(tempName) | FileAttributes.Hidden);
            }
            catch
            {
                // this doesn't have to succeed
            }
        }
        private void InstallFile(string sourceFileName)
        {
            var targetFileName = Path.Combine(this.TargetPath, Path.GetFileName(sourceFileName));
            this.Log($"Installing {sourceFileName} to {targetFileName}...");
            int retryCount = 0;

            Retry:
            try
            {
                File.Copy(sourceFileName, targetFileName);
            }
            catch (Exception ex)
            {
                if (retryCount < 5)
                {
                    this.Log($"Could not copy {sourceFileName}; will try again in 5 seconds. Error message: {ex.Message}");
                    Thread.Sleep(5000);
                    retryCount++;
                    goto Retry;
                }

                throw new Exception($"Could not install {targetFileName}. Error message: {ex.Message}");
            }

            this.installedFiles.Add(targetFileName);
        }

        private void Log(string message) => this.logWriter.WriteLine(message);
        private void LogError(string message) => this.logWriter.WriteLine("ERROR: " + message);
    }
}
